#include <avr/io.h>
 
#define THRESHOLD_HYSTERISIS 1
#define THRESHOLD_MIN 1
#define THRESHOLD_25MA 2
#define THRESHOLD_50MA 4
#define THRESHOLD_250MA 19
#define THRESHOLD_500MA 39
#define THRESHOLD_750MA 58
#define THRESHOLD_1500MA 116
 
#define LED_COMMON_PORTx PORTB
#define LED_COMMON_DDRx DDRB
#define LED_COLOUR_PORTx PORTD
#define LED_COLOUR_DDRx DDRD
#define LED_RED_BIT 5
#define LED_GREEN_BIT 6
#define LED_BLUE_BIT 7
#define LED_ALL_BITS (_BV(LED_RED_BIT) | _BV(LED_GREEN_BIT) | _BV(LED_BLUE_BIT))
 
uint8_t last_state[6] = { 0, 0, 0, 0, 0, 0, };
 
void adc_start(uint8_t chan)
{
    ADMUX = _BV(REFS0) | _BV(ADLAR) | (chan & 0x07); // setup reference, result alignment, and select channel
    ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADPS2); // enable and start ADC conversion with appropriate clock prescaler
}
 
uint8_t adc_read()
{
    loop_until_bit_is_set(ADCSRA, ADIF); // wait for conversion to finish
    ADCSRA |= _BV(ADIF); // reset the flag
    return ADCH; // read back (already aligned 8 bits)
}
 
void setup()
{
    // setup the pin directions
    PORTC = 0;
    DDRC = 0;
    LED_COMMON_PORTx = 0;
    LED_COMMON_DDRx |= 0x3F;
    LED_COLOUR_DDRx |= LED_ALL_BITS;
    adc_start(0); // start the first read
}
 
void loop()
{
    uint8_t i;
    for (i = 0; i < 6; i++)
    {
        uint8_t a = adc_read(i);
        adc_start((i + 1) % 6); // start the next read (this technique gives good timing because we wait less, timing means even LED brightness)
 
        uint8_t old_state = last_state[i];
        uint8_t new_state = 0;
         
        // apply hysterisis logic to combat noise (not very effective in real life but I tried)
        // TODO, add LPF filter or lower sample rate
        if (a == 0) {
            new_state = 0;
        }
        else if ((old_state == 4 && a > (THRESHOLD_1500MA - THRESHOLD_HYSTERISIS)) || (old_state == 3 && a > (THRESHOLD_1500MA + THRESHOLD_HYSTERISIS)) || (old_state < 3 && a > (THRESHOLD_1500MA - 0))) {
            new_state = 4;
        }
        else if ((old_state >= 4 && a <= (THRESHOLD_1500MA - THRESHOLD_HYSTERISIS)) || (old_state == 3 && a > (THRESHOLD_750MA - THRESHOLD_HYSTERISIS)) || (old_state == 2 && a > (THRESHOLD_750MA + THRESHOLD_HYSTERISIS)) || (old_state < 2 && a > (THRESHOLD_750MA + 0))) {
            new_state = 3;
        }
        else if ((old_state >= 3 && a <= (THRESHOLD_750MA - THRESHOLD_HYSTERISIS)) || (old_state == 2 && a > (THRESHOLD_250MA - THRESHOLD_HYSTERISIS)) || (old_state == 1 && a > (THRESHOLD_250MA + THRESHOLD_HYSTERISIS)) || (old_state < 1 && a > (THRESHOLD_250MA + 0))) {
            new_state = 2;
        }
        else if ((old_state >= 2 && a <= (THRESHOLD_250MA - THRESHOLD_HYSTERISIS)) || (old_state == 1 && a > (THRESHOLD_MIN - THRESHOLD_HYSTERISIS)) || (old_state == 0 && a > (THRESHOLD_MIN + THRESHOLD_HYSTERISIS))) {
            new_state = 1;
        }
         
        last_state[i] = new_state;
         
        if (new_state == 0)
        {
            // turn off LEDs
            LED_COMMON_PORTx = 0;
            LED_COLOUR_DDRx = 0;
        }
        else
        {
            // activate particular colour for particular index on LED matrix
            LED_COMMON_PORTx = _BV(i);
            if (new_state == 1) {
                LED_COLOUR_DDRx = _BV(LED_BLUE_BIT);
                LED_COLOUR_PORTx = LED_ALL_BITS & (~_BV(LED_BLUE_BIT));
            }
            else if (new_state == 2) {
                LED_COLOUR_DDRx = _BV(LED_GREEN_BIT);
                LED_COLOUR_PORTx = LED_ALL_BITS & (~_BV(LED_GREEN_BIT));
            }
            else if (new_state == 3) {
                LED_COLOUR_DDRx = _BV(LED_RED_BIT);
                LED_COLOUR_PORTx = LED_ALL_BITS & (~_BV(LED_RED_BIT));
            }
            else if (new_state >= 4) {
                LED_COLOUR_DDRx = _BV(LED_BLUE_BIT) | _BV(LED_RED_BIT);
                LED_COLOUR_PORTx = LED_ALL_BITS & (~(_BV(LED_BLUE_BIT) | _BV(LED_RED_BIT)));
            }
        }
    }
}
 
#ifndef ARDUINO
int main(void)
{
    setup();
    while(1) {
        loop();
    }
}
#endif